@@ -31,35 +31,47 @@ import contextlib
 from datetime import datetime
 from sqlalchemy import engine_from_config, event
 from sqlalchemy.orm import validates, scoped_session, sessionmaker
+from flask import has_app_context
 import modulemd as _modulemd
 
-from module_build_service import db, log, get_url_for
+from module_build_service import db, log, get_url_for, app, conf
 import module_build_service.messaging
 
 from sqlalchemy.orm import lazyload
 
 # Just like koji.BUILD_STATES, except our own codes for modules.
 BUILD_STATES = {
-    # When you parse the modulemd file and know the nvr and you create a
-    # record in the db, and that's it.
-    # publish the message
-    # validate that components are available
-    #   and that you can fetch them.
-    # if all is good, go to wait: telling module_build_service_daemon to take over.
-    # if something is bad, go straight to failed.
+    # This is (obviously) the first state a module build enters.
+    #
+    # When a user first submits a module build, it enters this state. We parse
+    # the modulemd file, learn the NVR, create a record for the module build.
+    # and publish the message.
+    #
+    # Then, we validate that the components are available, and that we can
+    # fetch them. If this is all good, then we set the build to the 'wait'
+    # state. If anything goes wrong, we jump immediately to the 'failed' state.
     "init": 0,
-    # Here, the scheduler picks up tasks in wait.
-    # switch to build immediately.
-    # throttling logic (when we write it) goes here.
+
+    # Here, the scheduler picks up tasks in wait and switches to build
+    # immediately. Eventually, we'll add throttling logic here so we don't
+    # submit too many builds for the build system to handle
     "wait": 1,
-    # Actively working on it.
+
+    # The scheduler works on builds in this state. We prepare the buildroot,
+    # submit builds for all the components, and wait for the results to come
+    # back.
     "build": 2,
-    # All is good
+
+    # Once all components have succeeded, we set the top-level module build
+    # to 'done'.
     "done": 3,
-    # Something failed
+
+    # If any of the component builds fail, then we set the top-level module
+    # build to 'failed' also.
     "failed": 4,
+
     # This is a state to be set when a module is ready to be part of a
-    # larger compose.  perhaps it is set by an external service that knows
+    # larger compose. perhaps it is set by an external service that knows
     # about the Grand Plan.
     "ready": 5,
 }
@@ -67,24 +79,58 @@ BUILD_STATES = {
 INVERSE_BUILD_STATES = {v: k for k, v in BUILD_STATES.items()}
 
 
+def _utc_datetime_to_iso(datetime_object):
+    """
+    Takes a UTC datetime object and returns an ISO formatted string
+    :param datetime_object: datetime.datetime
+    :return: string with datetime in ISO format
+    """
+    if datetime_object:
+        # Converts the datetime to ISO 8601
+        return datetime_object.strftime("%Y-%m-%dT%H:%M:%SZ")
+
+    return None
+
+
+@contextlib.contextmanager
+def _dummy_context_mgr():
+    """
+    Yields None. Used in the make_session to not duplicate the code when
+    app_context exists.
+    """
+    yield None
+
+
 @contextlib.contextmanager
 def make_session(conf):
-    # TODO - we could use ZopeTransactionExtension() here some day for
-    # improved safety on the backend.
-    engine = engine_from_config({
-        'sqlalchemy.url': conf.sqlalchemy_database_uri,
-    })
-    session = scoped_session(sessionmaker(bind=engine))()
-    event.listen(session, "before_commit", session_before_commit_handlers)
-    try:
-        yield session
-        session.commit()
-    except:
-        # This is a no-op if no transaction is in progress.
-        session.rollback()
-        raise
-    finally:
-        session.close()
+    """
+    Yields new SQLAlchemy database sesssion.
+    """
+    # Needs to be set to create app_context.
+    if 'SERVER_NAME' not in app.config or not app.config['SERVER_NAME']:
+        app.config['SERVER_NAME'] = 'localhost'
+
+    # If there is no app_context, we have to create one before creating
+    # the session. If we would create app_context after the session (this
+    # happens in get_url_for() method), new concurrent session would be
+    # created and this would lead to "database is locked" error for SQLite.
+    with app.app_context() if not has_app_context() else _dummy_context_mgr():
+        # TODO - we could use ZopeTransactionExtension() here some day for
+        # improved safety on the backend.
+        engine = engine_from_config({
+            'sqlalchemy.url': conf.sqlalchemy_database_uri,
+        })
+        session = scoped_session(sessionmaker(bind=engine))()
+        event.listen(session, "before_commit", session_before_commit_handlers)
+        try:
+            yield session
+            session.commit()
+        except Exception:
+            # This is a no-op if no transaction is in progress.
+            session.rollback()
+            raise
+        finally:
+            session.close()
 
 
 class MBSBase(db.Model):
@@ -102,6 +148,8 @@ class ModuleBuild(MBSBase):
     state_reason = db.Column(db.String)
     modulemd = db.Column(db.String, nullable=False)
     koji_tag = db.Column(db.String)  # This gets set after 'wait'
+    # Koji tag to which tag the Content Generator Koji build.
+    cg_build_koji_tag = db.Column(db.String)  # This gets set after wait
     copr_owner = db.Column(db.String)
     copr_project = db.Column(db.String)
     scmurl = db.Column(db.String)
@@ -109,19 +157,28 @@ class ModuleBuild(MBSBase):
     time_submitted = db.Column(db.DateTime, nullable=False)
     time_modified = db.Column(db.DateTime)
     time_completed = db.Column(db.DateTime)
+    new_repo_task_id = db.Column(db.Integer)
+    rebuild_strategy = db.Column(db.String, nullable=False)
 
     # A monotonically increasing integer that represents which batch or
     # iteration this module is currently on for successive rebuilds of its
     # components.  Think like 'mockchain --recurse'
     batch = db.Column(db.Integer, default=0)
 
+    rebuild_strategies = {
+        'all': 'All components will be rebuilt',
+        'changed-and-after': ('All components that have changed and those in subsequent batches '
+                              'will be rebuilt'),
+        'only-changed': 'All changed components will be rebuilt'
+    }
+
     def current_batch(self, state=None):
         """ Returns all components of this module in the current batch. """
 
         if not self.batch:
             raise ValueError("No batch is in progress: %r" % self.batch)
 
-        if state:
+        if state is not None:
             return [
                 component for component in self.component_builds
                 if component.batch == self.batch and component.state == state
@@ -132,11 +189,35 @@ class ModuleBuild(MBSBase):
                 if component.batch == self.batch
             ]
 
+    def last_batch_id(self):
+        """ Returns the id of the last batch """
+        return max([build.batch for build in self.component_builds])
+
+    def up_to_current_batch(self, state=None):
+        """
+        Returns all components of this module in the current batch and
+        in the previous batches.
+        """
+
+        if not self.batch:
+            raise ValueError("No batch is in progress: %r" % self.batch)
+
+        if state is not None:
+            return [
+                component for component in self.component_builds
+                if component.batch <= self.batch and component.state == state
+            ]
+        else:
+            return [
+                component for component in self.component_builds
+                if component.batch <= self.batch
+            ]
+
     def mmd(self):
         mmd = _modulemd.ModuleMetadata()
         try:
             mmd.loads(self.modulemd)
-        except:
+        except Exception:
             raise ValueError("Invalid modulemd")
         return mmd
 
@@ -148,6 +229,14 @@ class ModuleBuild(MBSBase):
             return BUILD_STATES[field]
         raise ValueError("%s: %s, not in %r" % (key, field, BUILD_STATES))
 
+    @validates('rebuild_strategy')
+    def validate_rebuild_stategy(self, key, rebuild_strategy):
+        if rebuild_strategy not in self.rebuild_strategies.keys():
+            choices = ', '.join(self.rebuild_strategies.keys())
+            raise ValueError('The rebuild_strategy of "{0}" is invalid. Choose from: {1}'
+                             .format(rebuild_strategy, choices))
+        return rebuild_strategy
+
     @classmethod
     def from_module_event(cls, session, event):
         if type(event) == module_build_service.messaging.MBSModule:
@@ -159,7 +248,7 @@ class ModuleBuild(MBSBase):
 
     @classmethod
     def create(cls, session, conf, name, stream, version, modulemd, scmurl, username,
-               copr_owner=None, copr_project=None):
+               copr_owner=None, copr_project=None, rebuild_strategy=None):
         now = datetime.utcnow()
         module = cls(
             name=name,
@@ -170,15 +259,18 @@ class ModuleBuild(MBSBase):
             scmurl=scmurl,
             owner=username,
             time_submitted=now,
+            time_modified=now,
             copr_owner=copr_owner,
             copr_project=copr_project,
+            # If the rebuild_strategy isn't specified, use the default
+            rebuild_strategy=rebuild_strategy or conf.rebuild_strategy
         )
         session.add(module)
         session.commit()
         module_build_service.messaging.publish(
             service='mbs',
             topic='module.state.change',
-            msg=module.json(),  # Note the state is "init" here...
+            msg=module.extended_json(),  # Note the state is "init" here...
             conf=conf,
         )
         return module
@@ -202,15 +294,44 @@ class ModuleBuild(MBSBase):
                                state_reason=state_reason)
         self.module_builds_trace.append(mbt)
 
-        log.debug("%r, state %r->%r" % (self, old_state, self.state))
+        log.info("%r, state %r->%r" % (self, old_state, self.state))
         if old_state != self.state:
             module_build_service.messaging.publish(
                 service='mbs',
                 topic='module.state.change',
-                msg=self.json(),  # Note the state is "init" here...
+                msg=self.extended_json(),
                 conf=conf,
             )
 
+    @classmethod
+    def local_modules(cls, session, name=None, stream=None):
+        """
+        Returns list of local module builds added by
+        utils.load_local_builds(...). When `name` or `stream` is set,
+        it is used to further limit the result set.
+
+        If conf.system is not set to "mock" or "test", returns empty
+        list everytime, because local modules make sense only when
+        building using Mock backend or during tests.
+        """
+        if conf.system in ["koji", "copr"]:
+            return []
+
+        filters = {}
+        if name:
+            filters["name"] = name
+        if stream:
+            filters["stream"] = stream
+        local_modules = session.query(ModuleBuild).filter_by(
+            **filters).all()
+        if not local_modules:
+            return []
+
+        local_modules = [m for m in local_modules
+                         if m.koji_tag and
+                         m.koji_tag.startswith(conf.mock_resultsdir)]
+        return local_modules
+
     @classmethod
     def by_state(cls, session, state):
         return session.query(ModuleBuild).filter_by(state=BUILD_STATES[state]).all()
@@ -222,7 +343,23 @@ class ModuleBuild(MBSBase):
 
         There should be at most one.
         """
-        tag = event.repo_tag.strip('-build')
+        if event.repo_tag.endswith('-build'):
+            tag = event.repo_tag[:-6]
+        else:
+            tag = event.repo_tag
+        query = session.query(cls)\
+            .filter(cls.koji_tag == tag)\
+            .filter(cls.state == BUILD_STATES["build"])
+
+        count = query.count()
+        if count > 1:
+            raise RuntimeError("%r module builds in flight for %r" % (count, tag))
+
+        return query.first()
+
+    @classmethod
+    def from_tag_change_event(cls, session, event):
+        tag = event.tag[:-6] if event.tag.endswith('-build') else event.tag
         query = session.query(cls)\
             .filter(cls.koji_tag == tag)\
             .filter(cls.state == BUILD_STATES["build"])
@@ -236,57 +373,38 @@ class ModuleBuild(MBSBase):
     def json(self):
         return {
             'id': self.id,
-            'name': self.name,
-            'stream': self.stream,
-            'version': self.version,
             'state': self.state,
             'state_name': INVERSE_BUILD_STATES[self.state],
             'state_reason': self.state_reason,
-            'state_url': get_url_for('module_build', id=self.id),
-            'scmurl': self.scmurl,
+            'stream': self.stream,
+            'version': self.version,
             'owner': self.owner,
-            'time_submitted': self.time_submitted,
-            'time_modified': self.time_modified,
-            'time_completed': self.time_completed,
-            "tasks": self.tasks(),
+            'name': self.name,
+            'rebuild_strategy': self.rebuild_strategy,
+            'scmurl': self.scmurl,
+            'time_submitted': _utc_datetime_to_iso(self.time_submitted),
+            'time_modified': _utc_datetime_to_iso(self.time_modified),
+            'time_completed': _utc_datetime_to_iso(self.time_completed),
+            'koji_tag': self.koji_tag,
+            'tasks': self.tasks(),
+        }
+
+    def extended_json(self):
+        json = self.json()
+        json.update({
             # TODO, show their entire .json() ?
             'component_builds': [build.id for build in self.component_builds],
             'modulemd': self.modulemd,
-            'state_trace': [{'time': record.state_time,
+            'state_trace': [{'time': _utc_datetime_to_iso(record.state_time),
                              'state': record.state,
                              'state_name': INVERSE_BUILD_STATES[record.state],
                              'reason': record.state_reason}
                             for record
-                            in self.state_trace(self.id)]
-        }
+                            in self.state_trace(self.id)],
+            'state_url': get_url_for('module_build', id=self.id)
+        })
 
-    @staticmethod
-    def _utc_datetime_to_iso(datetime_object):
-        """
-        Takes a UTC datetime object and returns an ISO formatted string
-        :param datetime_object: datetime.datetime
-        :return: string with datetime in ISO format
-        """
-        if datetime_object:
-            # Converts the datetime to ISO 8601
-            return datetime_object.strftime("%Y-%m-%dT%H:%M:%SZ")
-
-        return None
-
-    def api_json(self):
-
-        return {
-            "id": self.id,
-            "state": self.state,
-            'state_name': INVERSE_BUILD_STATES[self.state],
-            'state_reason': self.state_reason,
-            "owner": self.owner,
-            "name": self.name,
-            "time_submitted": self._utc_datetime_to_iso(self.time_submitted),
-            "time_modified": self._utc_datetime_to_iso(self.time_modified),
-            "time_completed": self._utc_datetime_to_iso(self.time_completed),
-            "tasks": self.tasks()
-        }
+        return json
 
     def tasks(self):
         """
@@ -294,9 +412,19 @@ class ModuleBuild(MBSBase):
         """
         tasks = dict()
         if self.id and self.state != 'init':
-
-            for build in ComponentBuild.query.filter_by(module_id=self.id).options(lazyload('module_build')).all():
-                tasks["%s/%s" % (build.format, build.package)] = "%s/%s" % (build.task_id, build.state)
+            for build in ComponentBuild.query\
+                    .filter_by(module_id=self.id)\
+                    .options(lazyload('module_build'))\
+                    .all():
+                tasks[build.format] = tasks.get(build.format, {})
+                tasks[build.format][build.package] = dict(
+                    task_id=build.task_id,
+                    state=build.state,
+                    state_reason=build.state_reason,
+                    nvr=build.nvr,
+                    # TODO -- it would be really nice from a UX PoV to get a
+                    # link to the remote task here.
+                )
 
         return tasks
 
@@ -305,9 +433,10 @@ class ModuleBuild(MBSBase):
             module_id=module_id).order_by(ModuleBuildTrace.state_time).all()
 
     def __repr__(self):
-        return "<ModuleBuild %s, stream=%s, version=%s, state %r, batch %r, state_reason %r>" % (
-            self.name, self.stream, self.version,
-            INVERSE_BUILD_STATES[self.state], self.batch, self.state_reason)
+        return (("<ModuleBuild %s, id=%d, stream=%s, version=%s, state %r,"
+                 " batch %r, state_reason %r>")
+                % (self.name, self.id, self.stream, self.version, INVERSE_BUILD_STATES[self.state],
+                   self.batch, self.state_reason))
 
 
 class ModuleBuildTrace(MBSBase):
@@ -324,7 +453,7 @@ class ModuleBuildTrace(MBSBase):
         retval = {
             'id': self.id,
             'module_id': self.module_id,
-            'state_time': self.state_time,
+            'state_time': _utc_datetime_to_iso(self.state_time),
             'state': self.state,
             'state_reason': self.state_reason,
         }
@@ -332,8 +461,8 @@ class ModuleBuildTrace(MBSBase):
         return retval
 
     def __repr__(self):
-        return "<ModuleBuildTrace %s, module_id: %s, state_time: %r, state: %s, state_reason: %s>" % (
-            self.id, self.module_id, self.state_time, self.state, self.state_reason)
+        return ("<ModuleBuildTrace %s, module_id: %s, state_time: %r, state: %s, state_reason: %s>"
+                % (self.id, self.module_id, self.state_time, self.state, self.state_reason))
 
 
 class ComponentBuild(MBSBase):
@@ -352,6 +481,13 @@ class ComponentBuild(MBSBase):
     state_reason = db.Column(db.String)
     # This stays as None until the build completes.
     nvr = db.Column(db.String)
+    # True when this component build is tagged into buildroot (-build tag).
+    tagged = db.Column(db.Boolean, default=False)
+    # True when this component build is tagged into final tag.
+    tagged_in_final = db.Column(db.Boolean, default=False)
+    # True when this component build is build-time only (should be tagged only
+    # to -build tag)
+    build_time_only = db.Column(db.Boolean, default=False)
 
     # A monotonically increasing integer that represents which batch or
     # iteration this *component* is currently in.  This relates to the owning
@@ -366,7 +502,7 @@ class ComponentBuild(MBSBase):
 
     @classmethod
     def from_component_event(cls, session, event):
-        if type(event) == module_build_service.messaging.KojiBuildChange:
+        if isinstance(event, module_build_service.messaging.KojiBuildChange):
             if event.module_build_id:
                 return session.query(cls).filter_by(
                     task_id=event.task_id, module_id=event.module_build_id)\
@@ -382,6 +518,10 @@ class ComponentBuild(MBSBase):
         return session.query(cls).filter_by(
             package=component_name, module_id=module_id).first()
 
+    def state_trace(self, component_id):
+        return ComponentBuildTrace.query.filter_by(
+            component_id=component_id).order_by(ComponentBuildTrace.state_time).all()
+
     def json(self):
         retval = {
             'id': self.id,
@@ -403,6 +543,19 @@ class ComponentBuild(MBSBase):
 
         return retval
 
+    def extended_json(self):
+        json = self.json()
+        json.update({
+            'state_trace': [{'time': _utc_datetime_to_iso(record.state_time),
+                             'state': record.state,
+                             'state_name': INVERSE_BUILD_STATES[record.state],
+                             'reason': record.state_reason}
+                            for record
+                            in self.state_trace(self.id)]
+        })
+
+        return json
+
     def __repr__(self):
         return "<ComponentBuild %s, %r, state: %r, task_id: %r, batch: %r, state_reason: %s>" % (
             self.package, self.module_id, self.state, self.task_id, self.batch, self.state_reason)
@@ -417,13 +570,14 @@ class ComponentBuildTrace(MBSBase):
     state_reason = db.Column(db.String, nullable=True)
     task_id = db.Column(db.Integer, nullable=True)
 
-    component_build = db.relationship('ComponentBuild', backref='component_builds_trace', lazy=False)
+    component_build = db.relationship('ComponentBuild', backref='component_builds_trace',
+                                      lazy=False)
 
     def json(self):
         retval = {
             'id': self.id,
             'component_id': self.component_id,
-            'state_time': self.state_time,
+            'state_time': _utc_datetime_to_iso(self.state_time),
             'state': self.state,
             'state_reason': self.state_reason,
             'task_id': self.task_id,
@@ -432,8 +586,9 @@ class ComponentBuildTrace(MBSBase):
         return retval
 
     def __repr__(self):
-        return "<ComponentBuildTrace %s, component_id: %s, state_time: %r, state: %s, state_reason: %s, task_id: %s>" % (
-            self.id, self.component_id, self.state_time, self.state, self.state_reason, self.task_id)
+        return ("<ComponentBuildTrace %s, component_id: %s, state_time: %r, state: %s,"
+                " state_reason: %s, task_id: %s>") % (self.id, self.component_id, self.state_time,
+                                                      self.state, self.state_reason, self.task_id)
 
 
 def session_before_commit_handlers(session):
