@@ -25,18 +25,26 @@ This class reads and processes messages from the message bus it is configured
 to use.
 """
 
-import koji
 import inspect
 import itertools
+
+try:
+    # python3
+    import queue
+except ImportError:
+    # python2
+    import Queue as queue
+
+import koji
 import fedmsg.consumers
 import moksha.hub
 
-from module_build_service.errors import ValidationError
 from module_build_service.utils import module_build_state_from_msg
 import module_build_service.messaging
 import module_build_service.scheduler.handlers.repos
 import module_build_service.scheduler.handlers.components
 import module_build_service.scheduler.handlers.modules
+import module_build_service.scheduler.handlers.tags
 from module_build_service import models, log, conf
 
 
@@ -44,19 +52,37 @@ class MBSConsumer(fedmsg.consumers.FedmsgConsumer):
     """ This is triggered by running fedmsg-hub. This class is responsible for
     ingesting and processing messages from the message bus.
     """
-    topic = ['{}.{}.'.format(pref.rstrip('.'), cat)
-             for pref, cat
-             in itertools.product(
-                conf.messaging_topic_prefix,
-                module_build_service.messaging._messaging_backends[conf.messaging]['services'])]
-    if not topic:
-        topic = '*'
-    log.debug('Setting topics: {}'.format(', '.join(topic)))
     config_key = 'mbsconsumer'
 
+    # It is set to the id of currently handled module build. It is used to
+    # group all the log messages associated with single module build to
+    # per module build log file.
+    current_module_build_id = None
+
     def __init__(self, hub):
+        # Topic setting needs to be done *before* the call to `super`.
+
+        backends = module_build_service.messaging._messaging_backends
+        prefixes = conf.messaging_topic_prefix  # This is a list.
+        services = backends[conf.messaging]['services']
+        suffix = backends[conf.messaging]['topic_suffix']
+        self.topic = [
+            '{}.{}{}'.format(prefix.rstrip('.'), category, suffix)
+            for prefix, category in itertools.product(prefixes, services)
+        ]
+        if not self.topic:
+            self.topic = '*'
+        log.debug('Setting topics: {}'.format(', '.join(self.topic)))
+
+        # The call to `super` takes action based on the setting of topics above
         super(MBSConsumer, self).__init__(hub)
 
+        # Our call to `super` above should have initialized an `incoming` queue
+        # for us.. but in certain test situations, it does not.  So here,
+        # establish a fake `incoming` queue.
+        if not hasattr(self, 'incoming'):
+            self.incoming = queue.Queue()
+
         # These two values are typically provided either by the unit tests or
         # by the local build command.  They are empty in the production environ
         self.stop_condition = hub.config.get('mbsconsumer.stop_condition')
@@ -84,7 +110,8 @@ class MBSConsumer(fedmsg.consumers.FedmsgConsumer):
             koji.BUILD_STATES["DELETED"]: NO_OP,
         }
         self.on_module_change = {
-            models.BUILD_STATES["init"]: NO_OP,
+            models.BUILD_STATES[
+                "init"]: module_build_service.scheduler.handlers.modules.init,
             models.BUILD_STATES[
                 "wait"]: module_build_service.scheduler.handlers.modules.wait,
             models.BUILD_STATES["build"]: NO_OP,
@@ -97,6 +124,7 @@ class MBSConsumer(fedmsg.consumers.FedmsgConsumer):
         }
         # Only one kind of repo change event, though...
         self.on_repo_change = module_build_service.scheduler.handlers.repos.done
+        self.on_tag_change = module_build_service.scheduler.handlers.tags.tagged
         self.sanity_check()
 
     def shutdown(self):
@@ -107,6 +135,12 @@ class MBSConsumer(fedmsg.consumers.FedmsgConsumer):
 
     def validate(self, message):
         if conf.messaging == 'fedmsg':
+            # If this is a faked internal message, don't bother.
+            if isinstance(message, module_build_service.messaging.BaseMessage):
+                log.info("Skipping crypto validation for %r" % message)
+                return
+            # Otherwise, if it is a real message from the network, pass it
+            # through crypto validation.
             super(MBSConsumer, self).validate(message)
 
     def consume(self, message):
@@ -120,7 +154,7 @@ class MBSConsumer(fedmsg.consumers.FedmsgConsumer):
         if isinstance(message, module_build_service.messaging.BaseMessage):
             msg = message
         else:
-            msg = self.get_abstracted_msg(message['body'])
+            msg = self.get_abstracted_msg(message)
 
         # Primary work is done here.
         try:
@@ -133,17 +167,15 @@ class MBSConsumer(fedmsg.consumers.FedmsgConsumer):
             self.shutdown()
 
     def get_abstracted_msg(self, message):
-        # Convert the message to an abstracted message
-        if conf.messaging == 'fedmsg':
-            msg = module_build_service.messaging.BaseMessage.from_fedmsg(
-                message['topic'], message)
-        elif conf.messaging == 'amq':
-            msg = module_build_service.messaging.BaseMessage.from_amq(
-                message['topic'], message)
+        parser = module_build_service.messaging._messaging_backends[conf.messaging].get('parser')
+        if parser:
+            try:
+                return parser.parse(message)
+            except module_build_service.messaging.IgnoreMessage:
+                pass
         else:
-            raise ValueError('The messaging format "{0}" is not supported'
+            raise ValueError('{0} backend does not define a message parser'
                              .format(conf.messaging))
-        return msg
 
     def sanity_check(self):
         """ On startup, make sure our implementation is sane. """
@@ -172,7 +204,7 @@ class MBSConsumer(fedmsg.consumers.FedmsgConsumer):
         build = None
 
         # Choose a handler for this message
-        if type(msg) == module_build_service.messaging.KojiBuildChange:
+        if isinstance(msg, module_build_service.messaging.KojiBuildChange):
             handler = self.on_build_change[msg.build_new_state]
             build = models.ComponentBuild.from_component_event(session, msg)
             if build:
@@ -180,6 +212,9 @@ class MBSConsumer(fedmsg.consumers.FedmsgConsumer):
         elif type(msg) == module_build_service.messaging.KojiRepoChange:
             handler = self.on_repo_change
             build = models.ModuleBuild.from_repo_done_event(session, msg)
+        elif type(msg) == module_build_service.messaging.KojiTagChange:
+            handler = self.on_tag_change
+            build = models.ModuleBuild.from_tag_change_event(session, msg)
         elif type(msg) == module_build_service.messaging.MBSModule:
             handler = self.on_module_change[module_build_state_from_msg(msg)]
             build = models.ModuleBuild.from_module_event(session, msg)
@@ -187,25 +222,30 @@ class MBSConsumer(fedmsg.consumers.FedmsgConsumer):
             log.debug("Unhandled message...")
             return
 
+        if not build:
+            log.debug("No module associated with msg {}".format(msg.msg_id))
+            return
+
+        MBSConsumer.current_module_build_id = build.id
+
         # Execute our chosen handler
         idx = "%s: %s, %s" % (handler.__name__, type(msg).__name__, msg.msg_id)
         if handler is self.NO_OP:
             log.debug("Handler is NO_OP: %s" % idx)
         else:
-            log.debug("Calling %s" % idx)
+            log.info("Calling %s" % idx)
+            further_work = []
             try:
-                if build:
-                    further_work = handler(conf, session, msg) or []
-                else:
-                    log.warn("There's no module associated with message ID {}"
-                             .format(msg.msg_id))
-            except ValidationError as e:
-                if build:
-                    build.transition(conf, state=models.BUILD_STATES['failed'],
-                                     state_reason=str(e))
+                further_work = handler(conf, session, msg) or []
+            except Exception as e:
                 msg = 'Could not process message handler. See the traceback.'
                 log.exception(msg)
-                session.commit()
+                session.rollback()
+                if build:
+                    session.refresh(build)
+                    build.transition(conf, state=models.BUILD_STATES['failed'],
+                                     state_reason=str(e))
+                    session.commit()
 
             log.debug("Done with %s" % idx)
 
@@ -219,6 +259,8 @@ class MBSConsumer(fedmsg.consumers.FedmsgConsumer):
                 log.info("  Scheduling faked event %r" % event)
                 self.incoming.put(event)
 
+        MBSConsumer.current_module_build_id = None
+
 
 def get_global_consumer():
     """ Return a handle to the active consumer object, if it exists. """
@@ -237,3 +279,11 @@ def work_queue_put(msg):
     """ Artificially put a message into the work queue of the consumer. """
     consumer = get_global_consumer()
     consumer.incoming.put(msg)
+
+
+def fake_repo_done_message(tag_name):
+    msg = module_build_service.messaging.KojiRepoChange(
+        msg_id='a faked internal message',
+        repo_tag=tag_name + "-build",
+    )
+    work_queue_put(msg)
