@@ -21,7 +21,7 @@
 # Written by Matt Prahl <mprahl@redhat.com
 
 import os
-import copy
+import tarfile
 import module_build_service
 
 from datetime import datetime, timedelta
@@ -29,24 +29,40 @@ from module_build_service import db
 from module_build_service.config import init_config
 from module_build_service.models import ModuleBuild, ComponentBuild, make_session
 import modulemd
-from module_build_service.utils import get_scm_url_re
 import module_build_service.pdc
 
+base_dir = os.path.dirname(__file__)
 app = module_build_service.app
 conf = init_config(app)
 
-def init_data():
+
+def uncompress_vcrpy_cassette():
+    cassette_dir = base_dir + '/vcr-request-data/'
+    if not os.path.exists(cassette_dir):
+        archive = cassette_dir.rstrip('/') + ".tar.gz"
+        with tarfile.open(archive, mode='r:gz') as t:
+            t.extractall(base_dir)
+
+
+uncompress_vcrpy_cassette()
+
+
+def clean_database():
     db.session.remove()
     db.drop_all()
     db.create_all()
-    db.session.commit()
+
+
+def init_data():
+    clean_database()
     for index in range(10):
         build_one = ModuleBuild()
         build_one.name = 'nginx'
         build_one.stream = '1'
         build_one.version = 2
         build_one.state = 3
-        build_one.modulemd = ''  # Skipping since no tests rely on it
+        with open(os.path.join(base_dir, "staged_data", "nginx_mmd.yaml")) as mmd:
+            build_one.modulemd = mmd.read()
         build_one.koji_tag = 'module-nginx-1.2'
         build_one.scmurl = ('git://pkgs.domain.local/modules/nginx?'
                             '#ba95886c7a443b36a9ce31abda1f9bef22f2f8c9')
@@ -59,6 +75,7 @@ def init_data():
             datetime(2016, 9, 3, 11, 25, 32) + timedelta(minutes=(index * 10))
         build_one.time_completed = \
             datetime(2016, 9, 3, 11, 25, 32) + timedelta(minutes=(index * 10))
+        build_one.rebuild_strategy = 'changed-and-after'
 
         component_one_build_one = ComponentBuild()
         component_one_build_one.package = 'nginx'
@@ -102,6 +119,7 @@ def init_data():
             datetime(2016, 9, 3, 12, 27, 19) + timedelta(minutes=(index * 10))
         build_two.time_completed = \
             datetime(2016, 9, 3, 11, 27, 19) + timedelta(minutes=(index * 10))
+        build_two.rebuild_strategy = 'changed-and-after'
 
         component_one_build_two = ComponentBuild()
         component_one_build_two.package = 'postgresql'
@@ -144,6 +162,7 @@ def init_data():
         build_three.time_modified = \
             datetime(2016, 9, 3, 12, 28, 40) + timedelta(minutes=(index * 10))
         build_three.time_completed = None
+        build_three.rebuild_strategy = 'changed-and-after'
 
         component_one_build_three = ComponentBuild()
         component_one_build_three.package = 'rubygem-rails'
@@ -183,11 +202,8 @@ def init_data():
             session.commit()
 
 
-def scheduler_init_data():
-    db.session.remove()
-    db.drop_all()
-    db.create_all()
-    db.session.commit()
+def scheduler_init_data(communicator_state=None):
+    clean_database()
 
     current_dir = os.path.dirname(__file__)
     star_command_yml_path = os.path.join(
@@ -209,6 +225,7 @@ def scheduler_init_data():
     build_one.owner = 'Buzz Lightyear'
     build_one.time_submitted = datetime(2016, 12, 9, 11, 23, 20)
     build_one.time_modified = datetime(2016, 12, 9, 11, 25, 32)
+    build_one.rebuild_strategy = 'changed-and-after'
 
     component_one_build_one = module_build_service.models.ComponentBuild()
     component_one_build_one.package = 'communicator'
@@ -217,7 +234,7 @@ def scheduler_init_data():
          '#da95886c8a443b36a9ce31abda1f9bed22f2f9c2')
     component_one_build_one.format = 'rpms'
     component_one_build_one.task_id = 12312345
-    component_one_build_one.state = None
+    component_one_build_one.state = communicator_state
     component_one_build_one.nvr = 'communicator-1.10.1-2.module_starcommand_1_3'
     component_one_build_one.batch = 2
     component_one_build_one.module_id = 1
@@ -242,11 +259,8 @@ def scheduler_init_data():
         session.commit()
 
 
-def test_resuse_component_init_data():
-    db.session.remove()
-    db.drop_all()
-    db.create_all()
-    db.session.commit()
+def test_reuse_component_init_data():
+    clean_database()
 
     current_dir = os.path.dirname(__file__)
     formatted_testmodule_yml_path = os.path.join(
@@ -268,6 +282,7 @@ def test_resuse_component_init_data():
     build_one.time_submitted = datetime(2017, 2, 15, 16, 8, 18)
     build_one.time_modified = datetime(2017, 2, 15, 16, 19, 35)
     build_one.time_completed = datetime(2017, 2, 15, 16, 19, 35)
+    build_one.rebuild_strategy = 'changed-and-after'
 
     component_one_build_one = module_build_service.models.ComponentBuild()
     component_one_build_one.package = 'perl-Tangerine'
@@ -340,6 +355,7 @@ def test_resuse_component_init_data():
     build_two.owner = 'Tom Brady'
     build_two.time_submitted = datetime(2017, 2, 19, 16, 8, 18)
     build_two.time_modified = datetime(2017, 2, 19, 16, 8, 18)
+    build_two.rebuild_strategy = 'changed-and-after'
 
     component_one_build_two = module_build_service.models.ComponentBuild()
     component_one_build_two.package = 'perl-Tangerine'
@@ -396,3 +412,110 @@ def test_resuse_component_init_data():
         session.add(component_three_build_two)
         session.add(component_four_build_two)
         session.commit()
+
+
+def test_reuse_shared_userspace_init_data():
+    clean_database()
+
+    with make_session(conf) as session:
+        mmd = modulemd.ModuleMetadata()
+
+        # Create shared-userspace-570, state is COMPLETE, all components
+        # are properly built.
+        current_dir = os.path.dirname(__file__)
+        formatted_testmodule_yml_path = os.path.join(
+            current_dir, 'staged_data', 'shared-userspace-570.yaml')
+        with open(formatted_testmodule_yml_path, 'r') as f:
+            yaml = f.read()
+            mmd.loads(yaml)
+
+        build_one = module_build_service.models.ModuleBuild()
+        build_one.name = mmd.name
+        build_one.stream = mmd.stream
+        build_one.version = mmd.version
+        build_one.state = 5
+        build_one.modulemd = yaml
+        build_one.koji_tag = 'module-testmodule-master-20170109091357'
+        build_one.scmurl = ('git://pkgs.stg.fedoraproject.org/modules/testmodule.'
+                            'git?#7fea453')
+        build_one.batch = 16
+        build_one.owner = 'Tom Brady'
+        build_one.time_submitted = datetime(2017, 2, 15, 16, 8, 18)
+        build_one.time_modified = datetime(2017, 2, 15, 16, 19, 35)
+        build_one.time_completed = datetime(2017, 2, 15, 16, 19, 35)
+        build_one.rebuild_strategy = 'changed-and-after'
+
+        session.add(build_one)
+
+        components = mmd.components.all
+        components.sort(key=lambda x: x.buildorder)
+        previous_buildorder = None
+        batch = 1
+        for pkg in components:
+            # Increment the batch number when buildorder increases.
+            if previous_buildorder != pkg.buildorder:
+                previous_buildorder = pkg.buildorder
+                batch += 1
+
+            pkgref = mmd.xmd['mbs']['rpms'][pkg.name]['ref']
+            full_url = pkg.repository + "?#" + pkgref
+            build = module_build_service.models.ComponentBuild(
+                module_id=1,
+                package=pkg.name,
+                format="rpms",
+                scmurl=full_url,
+                batch=batch,
+                ref=pkgref,
+                state=1
+            )
+            session.add(build)
+
+        # Create shared-userspace-577, state is WAIT, no component built
+        formatted_testmodule_yml_path = os.path.join(
+            current_dir, 'staged_data', 'shared-userspace-577.yaml')
+        with open(formatted_testmodule_yml_path, 'r') as f:
+            yaml = f.read()
+            mmd.loads(yaml)
+
+        build_one = module_build_service.models.ModuleBuild()
+        build_one.name = mmd.name
+        build_one.stream = mmd.stream
+        build_one.version = mmd.version
+        build_one.state = 3
+        build_one.modulemd = yaml
+        build_one.koji_tag = 'module-testmodule-master-20170109091357'
+        build_one.scmurl = ('git://pkgs.stg.fedoraproject.org/modules/testmodule.'
+                            'git?#7fea453')
+        build_one.batch = 0
+        build_one.owner = 'Tom Brady'
+        build_one.time_submitted = datetime(2017, 2, 15, 16, 8, 18)
+        build_one.time_modified = datetime(2017, 2, 15, 16, 19, 35)
+        build_one.time_completed = datetime(2017, 2, 15, 16, 19, 35)
+        build_one.rebuild_strategy = 'changed-and-after'
+
+        session.add(build_one)
+
+        components = mmd.components.all
+        # Store components to database in different order than for 570 to
+        # reproduce the reusing issue.
+        components.sort(key=lambda x: len(x.name))
+        components.sort(key=lambda x: x.buildorder)
+        previous_buildorder = None
+        batch = 1
+        for pkg in components:
+            # Increment the batch number when buildorder increases.
+            if previous_buildorder != pkg.buildorder:
+                previous_buildorder = pkg.buildorder
+                batch += 1
+
+            pkgref = mmd.xmd['mbs']['rpms'][pkg.name]['ref']
+            full_url = pkg.repository + "?#" + pkgref
+            build = module_build_service.models.ComponentBuild(
+                module_id=2,
+                package=pkg.name,
+                format="rpms",
+                scmurl=full_url,
+                batch=batch,
+                ref=pkgref
+            )
+            session.add(build)
